import argparse
import getpass
import sys

import requests

import spoc

APP_ERROR_STRINGS = {
    'AppAlreadyInstalledError': 'Application {} is already installed',
    'AppNotInstalledError': 'Application {} is not installed',
    'AppNotInRepoError': 'Application {} does not exist in the repository',
    'AppNotUpdateableError': 'Application {} does not have a newer version to update',
}

def handle_app_error(exception):
    ex_type = type(exception).__name__
    print(APP_ERROR_STRINGS[ex_type].format(exception.app_name), file=sys.stderr)

def handle_repo_error(exception):
    if isinstance(exception, requests.HTTPError):
        status_code = exception.response.status_code
        if status_code == 401:
            reason = 'Invalid username/password'
        else:
            reason = f'{status_code} {exception.response.reason}'
    else:
        ex_type = type(exception)
        reason = f'{ex_type.__module__}.{ex_type.__name__}'
    print(f'Repository "{spoc.config.REGISTRY_HOST}" cannot be reached due to: {reason}',
          file=sys.stderr)

def listing(list_type):
    if list_type == 'installed':
        apps = spoc.list_installed()
    elif list_type == 'online':
        apps = spoc.list_online()
    elif list_type == 'updates':
        apps = spoc.list_updates()
    else:
        apps = {}
    for app_name, app_version in apps.items():
        print(app_name, app_version)

def install(app_name, from_file):
    spoc.install(app_name, from_file)

def update(app_name, from_file):
    spoc.update(app_name, from_file)

def uninstall(app_name):
    spoc.uninstall(app_name)

def start(app_name):
    spoc.start(app_name)

def stop(app_name):
    spoc.stop(app_name)

def status(app_name):
    print(spoc.status(app_name))

def set_autostart(app_name, value):
    enabled = value.lower() in ('1', 'on', 'enable', 'true')
    spoc.set_autostart(app_name, enabled)

def start_autostarted():
    spoc.start_autostarted()

def stop_all():
    spoc.stop_all()

def login(host):
    username = input('Username: ')
    password = getpass.getpass()
    spoc.login(host, username, password)
    print('Login OK')

def prune():
    spoc.prune()


def parse_args(args=None):
    parser = argparse.ArgumentParser(description='SPOC application manager')
    subparsers = parser.add_subparsers(dest='action', required=True)

    parser_list = subparsers.add_parser('list')
    parser_list.set_defaults(action=listing)
    parser_list.add_argument('type', choices=('installed', 'online', 'updates'),
                             default='installed', const='installed', nargs='?',
                             help='Selected repository or application criteria')

    parser_install = subparsers.add_parser('install')
    parser_install.set_defaults(action=install)
    parser_install.add_argument('app', help='Name of the application to install')
    parser_install.add_argument('--from-file',
                                help='Filename containing the application definition ' \
                                     'to be used instead of online repository')

    parser_update = subparsers.add_parser('update')
    parser_update.set_defaults(action=update)
    parser_update.add_argument('app', help='Name of the application to update')
    parser_update.add_argument('--from-file',
                               help='Filename containing the application definition ' \
                                    'to be used instead of online repository')

    parser_uninstall = subparsers.add_parser('uninstall')
    parser_uninstall.set_defaults(action=uninstall)
    parser_uninstall.add_argument('app', help='Name of the application to uninstall')

    parser_start = subparsers.add_parser('start')
    parser_start.set_defaults(action=start)
    parser_start.add_argument('app', help='Name of the application to start')

    parser_stop = subparsers.add_parser('stop')
    parser_stop.set_defaults(action=stop)
    parser_stop.add_argument('app', help='Name of the application to stop')

    parser_status = subparsers.add_parser('status')
    parser_status.set_defaults(action=status)
    parser_status.add_argument('app', nargs='?', help='Name of the application to check')

    parser_autostart = subparsers.add_parser('autostart')
    parser_autostart.set_defaults(action=set_autostart)
    parser_autostart.add_argument('app', help='Name of the application to be automatically started')
    parser_autostart.add_argument('value', choices=('1', 'on', 'enable', 'true',
                                                    '0', 'off', 'disable', 'false'),
                                  help='Set or unset the applications to be automatically ' \
                                       'started after the host boots up')

    parser_start_autostarted = subparsers.add_parser('start-autostarted')
    parser_start_autostarted.set_defaults(action=start_autostarted)

    parser_stop_all = subparsers.add_parser('stop-all')
    parser_stop_all.set_defaults(action=stop_all)

    parser_login = subparsers.add_parser('login')
    parser_login.set_defaults(action=login)
    parser_login.add_argument('host', help='Hostname of the container registry')

    parser_prune = subparsers.add_parser('prune')
    parser_prune.set_defaults(action=prune)

    return parser.parse_args(args)

def main():  # pylint: disable=too-many-branches
    args = parse_args()
    try:
        if args.action is listing:
            listing(args.type)
        elif args.action is install:
            install(args.app, args.from_file)
        elif args.action is update:
            update(args.app, args.from_file)
        elif args.action is uninstall:
            uninstall(args.app)
        elif args.action is start:
            start(args.app)
        elif args.action is stop:
            stop(args.app)
        elif args.action is status:
            status(args.app)
        elif args.action is set_autostart:
            set_autostart(args.app, args.value)
        elif args.action is start_autostarted:
            start_autostarted()
        elif args.action is stop_all:
            stop_all()
        elif args.action is login:
            login(args.host)
        elif args.action is prune:
            prune()
    except spoc.AppError as ex:
        handle_app_error(ex)
    except requests.RequestException as ex:
        handle_repo_error(ex)
